/**************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*             Copyright SLF/WSL, 03/2007 - www.slf.ch                     *
**************************************************************************/
/* Here, we put all functions related to the details of reading a given data type */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>		/* for time calculations */
#include <math.h>
#include <inttypes.h>		/* C99 portable data types */
#include "limits.h"
#include "types_processing.h"

/* Functions for swapping byte order. */
/* Inspired from  http://www.gamedev.net/reference/articles/article2091.asp (GPL)*/
/* Itself inspired by Quake2 source code */

short int ReadInt2(frame_structure *frame) {
/* This function reads an int coded on 2 bytes */
/* It also updates the char index to point to the next byte */
	short int result;
	
	#ifdef BIGENDIAN
		result=(short int)((frame->raw[frame->cindex+0] << 8) + frame->raw[frame->cindex+1]);
	#else
		result=(short int)((frame->raw[frame->cindex+1] << 8) + frame->raw[frame->cindex+0]);
	#endif
	frame->cindex += 2;
	return result;
}


short int ReadInt2Swap(frame_structure *frame) {
/* This function reads an int coded on 2 bytes */
/* It also updates the char index to point to the next byte */
	short int result;
	
	#ifndef BIGENDIAN
		result=(short int)((frame->raw[frame->cindex+0] << 8) + frame->raw[frame->cindex+1]);
	#else
		result=(short int)((frame->raw[frame->cindex+1] << 8) + frame->raw[frame->cindex+0]);
	#endif
	frame->cindex += 2;
	return result;
}

int ReadInt4(frame_structure *frame) {
/* This function reads an int coded on 4 bytes */
/* It also updates the char index to point to the next byte */
	int result;
	
	#ifdef BIGENDIAN
		result=(int)( frame->raw[frame->cindex+3] + (frame->raw[frame->cindex+2] << 8) + (frame->raw[frame->cindex+1] << 16) + (frame->raw[frame->cindex+0] << 24) );
	#else
		result=(int)( frame->raw[frame->cindex+0] + (frame->raw[frame->cindex+1] << 8) + (frame->raw[frame->cindex+2] << 16) + (frame->raw[frame->cindex+3] << 24) );
	#endif
	frame->cindex += 4;
	
	return result;
}

int ReadInt4Swap(frame_structure *frame) {
/* This function reads an int coded on 4 bytes */
/* It also updates the char index to point to the next byte */
	int result;
	
	#ifdef BIGENDIAN
		result=(int)( frame->raw[frame->cindex+0] + (frame->raw[frame->cindex+1] << 8) + (frame->raw[frame->cindex+2] << 16) + (frame->raw[frame->cindex+3] << 24) );
	#else
		result=(int)( frame->raw[frame->cindex+3] + (frame->raw[frame->cindex+2] << 8) + (frame->raw[frame->cindex+1] << 16) + (frame->raw[frame->cindex+0] << 24) );
	#endif
	frame->cindex += 4;
	
	return result;
}

int64_t ReadInt8(frame_structure *frame) {
/* This function reads an int coded on 8 bytes */
/* It also updates the char index to point to the next byte */
	int64_t result;

	//an int is at least coded on 16 bits (C standard) -  a char is 8 bits
	#ifdef BIGENDIAN
		result=( frame->raw[frame->cindex+7] + (frame->raw[frame->cindex+6] << 8) + ((int64_t)frame->raw[frame->cindex+5] << 16) + ((int64_t)frame->raw[frame->cindex+4] << 24) + ((int64_t)frame->raw[frame->cindex+3] << 32) + ((int64_t)frame->raw[frame->cindex+2] << 40) + ((int64_t)frame->raw[frame->cindex+1] << 48) + ((int64_t)frame->raw[frame->cindex+0] << 56) );
	#else
		result=( frame->raw[frame->cindex+0] + (frame->raw[frame->cindex+1] << 8) + ((int64_t)frame->raw[frame->cindex+2] << 16) + ((int64_t)frame->raw[frame->cindex+3] << 24) + ((int64_t)frame->raw[frame->cindex+4] << 32) + ((int64_t)frame->raw[frame->cindex+5] << 40) + ((int64_t)frame->raw[frame->cindex+6] << 48) + ((int64_t)frame->raw[frame->cindex+7] << 56) );
	#endif
	frame->cindex += 8;

	return result;
}

uint64_t ReadUInt8(frame_structure *frame) {
/* This function reads an int coded on 8 bytes */
/* It also updates the char index to point to the next byte */
	uint64_t result;

	//an int is at least coded on 16 bits (C standard) -  a char is 8 bits
	#ifdef BIGENDIAN
		result=( frame->raw[frame->cindex+7] + (frame->raw[frame->cindex+6] << 8) + ((int64_t)frame->raw[frame->cindex+5] << 16) + ((int64_t)frame->raw[frame->cindex+4] << 24) + ((int64_t)frame->raw[frame->cindex+3] << 32) + ((int64_t)frame->raw[frame->cindex+2] << 40) + ((int64_t)frame->raw[frame->cindex+1] << 48) + ((int64_t)frame->raw[frame->cindex+0] << 56) );
	#else
		result=( frame->raw[frame->cindex+0] + (frame->raw[frame->cindex+1] << 8) + ((int64_t)frame->raw[frame->cindex+2] << 16) + ((int64_t)frame->raw[frame->cindex+3] << 24) + ((int64_t)frame->raw[frame->cindex+4] << 32) + ((int64_t)frame->raw[frame->cindex+5] << 40) + ((int64_t)frame->raw[frame->cindex+6] << 48) + ((int64_t)frame->raw[frame->cindex+7] << 56) );
	#endif
	frame->cindex += 8;

	return result;
}

float32_t ReadFloat4(frame_structure *frame) {
/* This function reads a float coded on 4 bytes */
/* It also updates the char index to point to the next byte */
	union
	{
		float32_t f;
		unsigned char b[4];
	} dat2;

	#ifndef BIGENDIAN
		dat2.b[0] = frame->raw[frame->cindex];
		dat2.b[1] = frame->raw[frame->cindex+1];
		dat2.b[2] = frame->raw[frame->cindex+2];
		dat2.b[3] = frame->raw[frame->cindex+3];
	#else
		dat2.b[0] = frame->raw[frame->cindex+3];
		dat2.b[1] = frame->raw[frame->cindex+2];
		dat2.b[2] = frame->raw[frame->cindex+1];
		dat2.b[3] = frame->raw[frame->cindex];
	#endif

	frame->cindex += 4;
	return dat2.f;
}

float32_t ReadFloat4Swap(frame_structure *frame) {
/* This function reads a float coded on 4 bytes */
/* It also updates the char index to point to the next byte */
	union
	{
		float32_t f;
		unsigned char b[4];
	} dat2;

	#ifdef BIGENDIAN
		dat2.b[0] = frame->raw[frame->cindex];
		dat2.b[1] = frame->raw[frame->cindex+1];
		dat2.b[2] = frame->raw[frame->cindex+2];
		dat2.b[3] = frame->raw[frame->cindex+3];
	#else
		dat2.b[0] = frame->raw[frame->cindex+3];
		dat2.b[1] = frame->raw[frame->cindex+2];
		dat2.b[2] = frame->raw[frame->cindex+1];
		dat2.b[3] = frame->raw[frame->cindex];
	#endif

	frame->cindex += 4;
	return dat2.f;
}

/////////////////////////////////////////////////////////////////////////////////////////
//////////////////// higher level functions, adapted to Cambell types ///////////////////
/////////////////////////////////////////////////////////////////////////////////////////

int Read_FP2(frame_structure *frame,const config_structure *config) {
	//2 bytes float 
	SIresult=(int16_t)ReadInt2Swap(frame);
	sign     = ((0x8000 & SIresult) >> 15 );
	exponent = ((0x6000 & SIresult) >> 13 );
	mantissa = ((0x1FFF & SIresult)       );
	switch(exponent) { 
		case 0: Fresult=(float)mantissa;
		break;
		case 1: Fresult=(float)mantissa*1e-1;
		break;
		case 2: Fresult=(float)mantissa*1e-2;
		break;
		default: Fresult=(float)mantissa*1e-3;
	}
	if(Fresult>=frame->fp2_nan || isnan(Fresult))
		fprintf(config->output,config->NAN,config->separator);
	else {
		if(sign ==0) 
			fprintf(config->output,config->fp2_format,config->separator,Fresult);
		else
			fprintf(config->output,config->fp2_format,config->separator,-Fresult);
	}

	return EXIT_SUCCESS;
}

int Read_IEEE4(frame_structure *frame,const config_structure *config) {
	//IEEE4 & L -> Little endian
	Fresult=ReadFloat4(frame);
	if(isnan(Fresult))
		fprintf(config->output,config->NAN,config->separator);
	else
		fprintf(config->output,config->floats_format,config->separator,Fresult);
	
	return EXIT_SUCCESS;
}

int Read_IEEE4B(frame_structure *frame,const config_structure *config) {
	//Big endian version
	Fresult=ReadFloat4Swap(frame);
	if(isnan(Fresult))
		fprintf(config->output,config->NAN,config->separator);
	else
		fprintf(config->output,config->floats_format,config->separator,Fresult);
	
	return EXIT_SUCCESS;
}

int Read_FP4(frame_structure *frame,const config_structure *config) {
	//in progress... but it should work! see Appendix C of CR10X manual TODO
	Iresult=ReadInt4(frame);
	sign     = ((0x80000000 & Iresult) >> 31 );
	exponent = ((0x7F000000 & Iresult) >> 24 );//isn't it 23 instead?
	mantissa = ((0x00FFFFFF & Iresult)       );
	Fresult=(float)mantissa/16777216.0*(float)pow(2.0,(double)(exponent-64));//TODO: optimize...(exponent is INT)
#ifdef DEBUG
	fprintf(stderr,"0x%0x->",Iresult);
	fprintf(stderr,"(%d)_%d*10^(%d)->%g\n",sign,mantissa,exponent,Fresult);
#endif
	fprintf(config->output,"%s",config->separator);
	/*if(Fresult>=frame->fp4_nan || isnan(Fresult))
		fprintf(config->output,config->NAN,config->separator);
	else {
		if(sign ==0) 
			fprintf(config->output,config->fp4_format,config->separator,Fresult);
		else
			fprintf(config->output,config->fp4_format,config->separator,-Fresult);
}*/

	return EXIT_SUCCESS;
}

int Read_UINT2(frame_structure *frame,const config_structure *config) {
	//2 bytes unsigned int
	SIresult=(int16_t)ReadInt2Swap(frame);
	if(SIresult>=frame->uint2_nan || isnan(SIresult))
		fprintf(config->output,config->NAN,config->separator);
	else {
		fprintf(config->output,config->ints_format,config->separator,SIresult);
	}
	
	return EXIT_SUCCESS;
}

int Read_INT4(frame_structure *frame,const config_structure *config) {
	//4 bytes int
	fprintf(config->output,config->ints_format,config->separator,ReadInt4Swap(frame));
	
	return EXIT_SUCCESS;
}

int Read_BOOL4(frame_structure *frame,const config_structure *config) {
	//4 bytes boolean: either FFFFFFFF or 00000000
	if(ReadInt4Swap(frame)==0)
		fprintf(config->output,config->bool_false,config->separator);
	else
		fprintf(config->output,config->bool_true,config->separator);
	
	return EXIT_SUCCESS;
}

int Read_ULONG(frame_structure *frame,const config_structure *config) {
	//TODO: check
	fprintf(config->output,config->lints_format,config->separator,ReadUInt8(frame));
	
	return EXIT_SUCCESS;
}

int Read_LONG(frame_structure *frame,const config_structure *config) {
	//TODO: check
	fprintf(config->output,config->lints_format,config->separator,ReadInt8(frame));
	
	return EXIT_SUCCESS;
}

int Read_NSec(frame_structure *frame,const config_structure *config) {
	//This is a timestamp: do we represent it a a raw number of seconds?
	time_offset = ReadInt4Swap(frame)+TO_EPOCH;
	strftime(time_str,MAX_FIELD,config->time_format,gmtime(&time_offset));
	fprintf(config->output,config->nsec_format,config->separator,time_str);
	
	return EXIT_SUCCESS;
}

int Read_ASCII(frame_structure *frame,const config_structure *config) {
	//this is for the ASCII strings
	count=frame->field_options[field_index];
	fprintf(config->output,config->strings_beg,config->separator);
	for(;count>0;count--) {
		Cresult=frame->raw[frame->cindex++];
		if(Cresult == 0) {//the end of the string has been reached
			frame->cindex += count-1;
			count=0;
		} else 
			fprintf(config->output,"%c",Cresult);
	}
	fprintf(config->output,config->strings_end,config->separator);
	
	return EXIT_SUCCESS;
}

